#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <fftw3.h>


void pitchdetect_confid(float *SPEC1,int N_pks, int L_freq,int L_start,int L_end,int N_pnts, float *PF_idx, float *PF_confid,int *PF_idxN1);

/*********************************
//Pitch detection algotithm Assumes that the pitch range is from 100Hz to 350Hz
//SPEC1 	-> Pointer to the array having the spectrum 
//L_freq	-> Total frequency to be used.
//L_start	-> First pitch frequency to be checked.
//L_end		-> Last pitch frequency to be checked.
//PF_idx	-> The array where the pitch is going to be stored.
//PF_confiid	-> The array confidance of estimating the pitchs in PF_idx.
//PF_idxN1	-> Number of total pitch frequency to be estimated.
********************************/
void pitchdetect_confid(float *SPEC1,int N_pks, int L_freq,int L_start,int L_end,int N_pnts, float *PF_idx, float *PF_confid,int *PF_idxN1)
  {
//  int N_pnts=1;     //N_pnts is the number of succesive harmonics to be checked.
  int N_deci=10;    //
  int FLG=0;
  int N_blks,m,blk,n,l,MAX_ID_PTCH,pks,ptc;
  float ptch_m,FB_O_tmp,n_chk,FUNDA_COST1_max=0,FUNDA_COST_tmp1,x;
  float *FUNDA_COST1,*PEAKS;
  int *IDX_PTCH, *ID_PK_PTCH;


  N_pnts=((N_pks-1)<N_pnts) ?  (N_pks-1):N_pnts;       //Number of peaks simultaneously considered is made same as 
  *PF_idxN1=0;
  N_blks=(int)floorf((float)(L_freq/2)/(float)L_end)-N_pnts;
  FUNDA_COST1=(float *)calloc(((L_end-L_start)*N_deci+L_start),sizeof(float));
  for(m=(L_start-1);m<((L_end-L_start)*N_deci+L_start);m++)
    {
    ptch_m=((float)(m-L_start+1))/((float)N_deci)+(float)L_start;
    for(blk=0;blk<N_blks;blk++)
      {
      FUNDA_COST_tmp1=1;
      n_chk=((blk+1)*ptch_m-1);
      while(n_chk<((blk+1+N_pnts)*ptch_m))
	{
	/*printf("m=%d,\t blk=%d,\t n_chk=%f\n",m,blk,n_chk);*/
        FUNDA_COST_tmp1*=SPEC1[(int)rint(n_chk)];
        n_chk+=ptch_m;
	}
      if(FUNDA_COST_tmp1>FUNDA_COST1[m])
        FUNDA_COST1[m]=FUNDA_COST_tmp1;
      }
    if(FUNDA_COST1[m]>FUNDA_COST1_max)
      FUNDA_COST1_max=FUNDA_COST1[m];
    }
  FUNDA_COST1_max*=0.3;   //Scaling for saving multiplication in later use.
  for(m=(L_start-1);m<((L_end-L_start)*N_deci+L_start);m++)
    {
    FUNDA_COST1[m]=((FUNDA_COST1[m]>FUNDA_COST1_max)? FUNDA_COST1[m]:0 );   //Thresholding for appreciable peaks, mic1.
    }
  ID_PK_PTCH=(int *)calloc(30,sizeof(int));  //Allocating space for Index of the peaks from peak_pick_win
  
    //Pitch for mic output 1.
  peak_pick_win(FUNDA_COST1,((L_end-L_start)*N_deci+L_start),30,ID_PK_PTCH,&MAX_ID_PTCH); //Picking the peaks in the detection function
  PEAKS=(float *)calloc(MAX_ID_PTCH,sizeof(float));  //For storing peak values
  if(MAX_ID_PTCH>0)
    {
    IDX_PTCH=(int *)calloc(MAX_ID_PTCH,sizeof(int));
    for(m=0;m<MAX_ID_PTCH;m++)     //Creating the array of peak values
      {
      PEAKS[m]=(float)ID_PK_PTCH[m];
      }
    sort(PEAKS, IDX_PTCH, MAX_ID_PTCH);    //Sorting the peaks 
    for(pks=0;pks<(MAX_ID_PTCH-1);pks++)
      for(m=pks+1;m<MAX_ID_PTCH;m++)
	if(PEAKS[pks] != 0)
	  {
	  x=rint(((PEAKS[m]-(float)L_start)/(float)N_deci+(float)L_start)/((PEAKS[pks]-(float)L_start)/(float)N_deci+(float)L_start));
	  if(fabs(x*((PEAKS[pks]-(float)L_start)/(float)N_deci+(float)L_start)-((PEAKS[m]-(float)L_start)/(float)N_deci+(float)L_start))<=3)
	    if(FUNDA_COST1[(int)PEAKS[pks]] > 1.1*FUNDA_COST1[(int)PEAKS[m]])
	      PEAKS[m]=PEAKS[pks];
	    else
	      PEAKS[pks]=PEAKS[m];
	  }
    //PF_idx=(float *)calloc(MAX_ID_PTCH,sizeof(float));
    //Making set difference and storing
    ptc=0;
    for(pks=0;pks<(MAX_ID_PTCH-1);pks++)
      {
      if(PEAKS[pks]>0)
	{
	if(ptc==0)
	  {
	  PF_idx[0]=(float)(PEAKS[pks]-L_start)/(float)N_deci+L_start;
          PF_confid[0]=FUNDA_COST1[(int)PEAKS[pks]];
          ptc++;
	  FLG=1;
	  }
        else
	  {
          for(m=0;m<ptc;m++)
	    {
	    if(fabsf(((float)(PEAKS[pks]-L_start)/(float)N_deci+L_start)-PF_idx[m])<0.1)
	      {
	      FLG=1;
              break;
	      }
	    }
	  }
	}
      else
	{
	FLG=1;
	}
      if(FLG==0)
	{
        PF_idx[ptc]=(float)(PEAKS[pks]-L_start)/(float)N_deci+L_start;
        PF_confid[ptc]=FUNDA_COST1[(int)PEAKS[pks]];
	ptc++;
	}
      FLG=0;
      }
    *PF_idxN1=ptc;
    free(IDX_PTCH);
    IDX_PTCH=NULL;
    free(PEAKS);
    PEAKS=NULL;
    }

  free(ID_PK_PTCH);
  free(FUNDA_COST1);
  }
