#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <fftw3.h>

void sigsep_final();   //Separate signals with more than 2 sensors.

//*******************************************************
//Signal separation algorithm implementation in C.
//Modified version of sigsep.c to function without using
//PITCH and VOICED UNVOICEd information in the weight decision. 
//The head is in picture and hence Inter aural attenuation difference 
// is used in estimating DOA.
//*******************************************************
void sigsep_final()
  {
  extern FL_FORM  *audio_processer;      //The form structure
  extern signal Data;                    //The structure having the data being processed.
  extern source *SOURCE;                 //The array of structures where the data will be stored.
  source *SOURCE_tmp;                    //The temporary array of structures where the data will be stored.
  extern int N_SOURCE;                   //Number of sources detected at any time.
  char* source_display_window_label;     //For making the string to be displayed as window name.
  int N_pad,LENGTH,L,L_pad, L_DOA_start, L_DOA_end, L_proc,L_DOAs,N_DOA_hist_bin,l; 		//Number of sensors.
  float micspacing,c,tmpi,MAX_TMP,TMP_VAL,SCALE_LR,VAL_DOA_end, VAL_DOA_start,DOA_end;
  int i,col,row,n,rcnt,ccnt,k,q,n_src,doa,doa_tmp=0,ptch,ptch_tmp,src,n_tmp,mic,tmp_src,src_tmp,m,n_src_tmp;   //Index variables
  matrix MEAN,TMP;
  fftw_complex *TMP1, *TMP2;                 //The signal matrix fromed from Data.channels in fftw format.
  int M=2,N_DOAS,N_DOAS_tmp,N_SOURCE_tmp, TMP_cnt, WEIGHT_SP_SUM[SOURCE_NUM];	//Number of mics
  int *ID_PK;
  float *DOAs, *DOA_full, *DOA_list, *DOA_list_tmp, *DOA_wt_hist, *DOA_hist_idx, *WGHT_TMP;
  float *SPEC1,*SPEC2, *SPEC, *DOA_AT_MAP;
  float **WEIGHT_SP;               //The array for storing pitch frequencies
  FILE*  fid;
  cplxnum ***FB_ORIG;           //The matrix of FFTs of the signals.
  fftw_plan FPLN;		//FFTW plan for the comptation
 // fftw_complex *TMP2;
  SCALE_LR=0.4/0.52;           //Scaling to compensate for unequal gain in the channels (VRIABLE IMPORTANT)

  if (Data.n_channels==1)
    {
    fl_show_messages("This is a single channel data. I need two ears for now!!");
    return;
    }

  printf("Start \n");
  SOURCE=(source *)calloc(SOURCE_NUM,sizeof(source)); //Reseving spacet to store the sources.
  SOURCE_tmp=(source *)calloc(SOURCE_NUM,sizeof(source)); //Reseving spacet to store the sources temporarily.
  c=331; 		//Velocity of sound.
  micspacing=0.17/(c/Data.fs);          //Mic sapcing in terms of  wavelength at sampling frequency.  //VARIABLE here 17cm (with head)
//  micspacing=0.135/(c/Data.fs);          //Mic sapcing in terms of  wavelength at sampling frequency.  //VARIABLE here 13cm
//  micspacing=0.0825/(c/Data.fs);          //Mic sapcing in terms of  wavelength at sampling frequency.  //VARIABLE here 8.25cm
  L=(int)rintf(1024*Data.fs/44100);
  L_pad=4096-L;              //Number of zeros to be padded.

   //Setting up structure for doing DOA detection by relative attenuation.
  //L_DOA_start=((L+L_pad)/(2*micspacing));  //Staring frequency from which DOA is to be estimated using attenuation
  L_DOA_start=(((L+L_pad)/(2*micspacing)))/2;   //Staring frequency from which DOA is to be estimated using attenuation, less ththan corresponding to wavelength/2
  VAL_DOA_start=2.5;
  L_DOA_end=1000;           //Ending frequency to which DOA is to be estimated using attenuation
  VAL_DOA_end=3;           
  DOA_end=40;               //DOA start by default is 90 degrees.
  DOA_AT_MAP=(float *)calloc(L+L_pad,sizeof(float)); //For storing the scale and corresonding DOA
  for(n=L_DOA_start;n<L_DOA_end;n++)  //Storing the map from 90 to 40 in terms of attenuation 
    DOA_AT_MAP[n]=(n-1)*(VAL_DOA_end-VAL_DOA_start)/(L_DOA_end-L_DOA_start)+VAL_DOA_start;

  //Creating the fft filterbank o/p  matrix
  FB_ORIG = (cplxnum ***) calloc(L,sizeof(cplxnum**));
  for(ccnt =0;ccnt<L;ccnt++)              //Up to length
    {
    FB_ORIG[ccnt]=(cplxnum **) calloc(M,sizeof(cplxnum*));         //Upto number of sensors
    for(rcnt = 0;rcnt<M; rcnt++)
      FB_ORIG[ccnt][rcnt] = (cplxnum *) calloc(L_pad+L,sizeof(cplxnum));       
    } 

  //Create FFTW plan for the computation
  TMP1 = fftw_malloc(sizeof(fftw_complex) * (L+L_pad));
  TMP2 = fftw_malloc(sizeof(fftw_complex) * (L+L_pad));
  FPLN = fftw_plan_dft_1d(L+L_pad,TMP1,TMP2,FFTW_FORWARD,FFTW_ESTIMATE);
  
  //Reserving space for spectrum
  SPEC1 = (float *) calloc(L_pad+L,sizeof(float));      //For storing spectral magnitude of channel 1.  
  SPEC2 = (float *) calloc(L_pad+L,sizeof(float));       //For storing spectral magnitude of channel 2.
  SPEC = (float *) calloc(L_pad+L,sizeof(float));        //For storing mean spectral magnitude.
  WGHT_TMP = (float *) calloc(L_pad+L,sizeof(float));       
  WEIGHT_SP=(float **) calloc(SOURCE_NUM,sizeof(float *));
  for(m=0;m<SOURCE_NUM;m++)
    {
    WEIGHT_SP[m]=(float *) calloc(L_pad+L,sizeof(float));
    }
  DOA_full=(float *) calloc(L_pad+L,sizeof(float));
  
  LENGTH=Data.channels.col-L;                  //Total length to be processed
  N_DOA_hist_bin=20;                      //Number of bins used in finding histogram of DOA estimates.  (VARIABLE)
         //Initialising the SOURCE structure
  SOURCE[0].DOA_SET=0;
  SOURCE[0].WT=0;             //For activity detection
  SOURCE[0].WTDOA=0;          //For DOA   
  N_SOURCE=0;          //Number of sources

                            //Allocating memories for DOA storage and histograms
  DOAs=(float *) calloc((int)(L+L_pad),sizeof(float));
  DOA_wt_hist=(float *) calloc(N_DOA_hist_bin,sizeof(float));
  DOA_hist_idx=(float *) calloc(N_DOA_hist_bin,sizeof(float));
         //Outer loop over time index. Most of the algo is within it.
  for(l=0;l<LENGTH;l++)
    {
    for(n=0;n<L;n++)        //Filling data in zeropadded array
      {
      TMP1[n][0]=Data.channels.m[0][l+n]*SCALE_LR;
      TMP1[n][1]=0;
      }
    for(n=L;n<(L+L_pad);n++)        //Filling zeros in zeropadded array
      {
      TMP1[n][0]=0;
      TMP1[n][1]=0;
      }
    fftw_execute(FPLN);               //Doing the fft
    for(n=0;n<(L+L_pad);n++)       //Filling the FB_ORIG array, channel1, using fftw output
      {
      FB_ORIG[(l % L)][0][n].re=TMP2[n][0];
      FB_ORIG[(l % L)][0][n].im=TMP2[n][1];
      }

    for(n=0;n<L;n++)        //Filling data in zeropadded array
      {
      TMP1[n][0]=Data.channels.m[1][l+n];
      TMP1[n][1]=0;
      }
    for(n=L;n<(L+L_pad);n++)        //Filling zeros in zeropadded array
      {
      TMP1[n][0]=0;
      TMP1[n][1]=0;
      }
    fftw_execute(FPLN);               //Doing the fft
    for(n=0;n<(L+L_pad);n++)       //Filling the FB_ORIG array, channel2  using fftw output
      {
      FB_ORIG[(l % L)][1][n].re=TMP2[n][0];
      FB_ORIG[(l % L)][1][n].im=TMP2[n][1];
      }

    if ((l%L)==(L-1))
      {
      fl_check_forms();           //So that form remains interactive or responsive.
      TMP_cnt=(int)rint((double)L/2);
      for (i=0;i<(L+L_pad);i++)   //To find abs spectrum at one of the frequencies.
	{
  	SPEC1[i]=(float)sqrt((double)(FB_ORIG[TMP_cnt][0][i].re*FB_ORIG[TMP_cnt][0][i].re+FB_ORIG[TMP_cnt][0][i].im*FB_ORIG[TMP_cnt][0][i].im));  //Power in channel 1
  	SPEC2[i]=(float)sqrt((double)(FB_ORIG[TMP_cnt][1][i].re*FB_ORIG[TMP_cnt][1][i].re+FB_ORIG[TMP_cnt][1][i].im*FB_ORIG[TMP_cnt][1][i].im));  //Power in channel 2
        SPEC[i]=(SPEC1[i]+SPEC2[i])/2;        //Mean power
	}

      for(m=1;m< L+L_pad;m++)           //Reinitialising DOAs storage to zero.
        DOAs[m]=0;
      //DOA_estims_all_scaled(FB_ORIG, SPEC1, SPEC2, DOAs, Data.fs, micspacing, c, L+L_pad, L);     //Estimating the DOAs in the lower freq range
      DOA_estims_all(FB_ORIG, DOAs, Data.fs, micspacing, c, L+L_pad, L);     //Estimating the DOAs in the lower freq range
      DOA_estims_atten(DOAs, SPEC1, SPEC2, DOA_AT_MAP, DOA_end, L_DOA_start, L_DOA_end);     //Atteneation based DOA estimation when head is used.
      L_DOAs=(int)((L+L_pad)/(2*micspacing));                        //Toatal number of sources detected.
      for(m=1;m< L_DOA_end;m++)   //Initiallising the compete DOA matrix over all frequency bins.
	{
        DOA_full[m]=DOAs[m];
	DOA_full[(L+L_pad)-m]=DOAs[m];
	}
      scale_hist(DOAs,SPEC,N_DOA_hist_bin,L_DOA_end,DOA_wt_hist,DOA_hist_idx); //Scaled histogram of DOAs is in HIST_val
      ID_PK=(int *)calloc(4,sizeof(int));  //Allocating space for Index of the peaks from peak_pick_win
      MAX_TMP=DOA_wt_hist[max_vec(DOA_wt_hist,N_DOA_hist_bin)];  //Maximum of the 
      for (i=0;i<N_DOA_hist_bin;i++)
	DOA_wt_hist[i]=((DOA_wt_hist[i]>(0.3*MAX_TMP)) ? DOA_wt_hist[i]:0);
      N_DOAS=0;     //Number of distinct DOAs detected as sources.
      peak_pick_win(DOA_wt_hist,N_DOA_hist_bin,4,ID_PK,&N_DOAS); //Picking the peaks in the weight DOA histogram
      DOA_list=(float *)calloc(N_DOAS,sizeof(float));        //Allocating space to store the DOAs of sources in this block
      i=0;
      while(i<N_DOAS)               //store the DOAs of sources in this block
        {
        if((DOA_hist_idx[ID_PK[i]]!=0) && (DOA_hist_idx[ID_PK[i]]!=180))
        DOA_list[i]=DOA_hist_idx[ID_PK[i]];
        i++;
        }
      N_DOAS=i;      

      if((N_DOAS>=1) && (N_SOURCE==0) )
        {
        SOURCE[0].DOA=DOA_list[0]; //Start a new data structure for that source.
	SOURCE[0].DOA_SET=1;         //Setting flag saying DOA has been allocated.
	N_SOURCE=1;                 //Incrementing the source count
        SOURCE[0].sig=(float *)calloc(LENGTH,sizeof(float));            //Initialize  the signal part as empty.
	SOURCE[0].L_src=0;         //Length of source signal after initialising.
	SOURCE[0].L_init=(l+1)-L;    //Length of source signal after initialising.
        SOURCE[0].WT=0;         //Initialise the weighting/flag for inclusion in this cycle. Mainly to determine if active
        SOURCE[0].WTDOA=1;      //Initialise the weighting/flag for inclusion in this cycle. Mainly to determine if DOA is to be used in this block.
	}  

      //Reducing(!) the source weight (Corresponding to decay of exciataion)
      for(n_src=0;n_src<N_SOURCE;n_src++)
	{
	SOURCE[n_src].WT++;
	SOURCE[n_src].WTDOA++;
	}
       
      //Assigning Section

      //Eliminating existing clues
      if (N_SOURCE>0)
	{
        DOA_list_tmp=(float *)calloc(N_DOAS,(float)sizeof(float));  
	for(src=0;src<N_SOURCE;src++)
	  {
	  if(N_DOAS>0)
	    {
	    for(n_tmp=0;n_tmp<N_DOAS;n_tmp++)
	      {
          //    printf("\nDOA_list[%d]=%f\n",n_tmp,DOA_list[n_tmp]);   //For checking DOA 
              DOA_list_tmp[n_tmp]=DOA_list[n_tmp];
              }
	    N_DOAS_tmp=N_DOAS;
	    N_DOAS=0;
	    for(doa=0;doa<N_DOAS_tmp;doa++)
	      {
	 //     if(SOURCE[src].DOA_SET==1)
		{
		if (fabs((double)SOURCE[src].DOA-(double)DOA_list[doa])<15)   //Was 15 on 11-11-2003
		  {
		  SOURCE[src].DOA=0.9*SOURCE[src].DOA+(1-0.9)*DOA_list[doa];    //Was originally 0.993
		  SOURCE[src].WTDOA=1;
                  SOURCE[src].WT=1;               //Added on 12-11-2003
		  }
		else
		  {
		  DOA_list_tmp[N_DOAS]=DOA_list[doa];
		  N_DOAS++;
		  }
		}
	      }
	    for(n_tmp=0;n_tmp<N_DOAS;n_tmp++)    //Impementing DOA_list=DOA_list_tmp;
              {
	      DOA_list[n_tmp]=DOA_list_tmp[n_tmp];
              }
	    } 
	  }
        free(DOA_list_tmp);
        DOA_list_tmp=NULL;
	}
              //New sources added to SORUCE
      for(doa=0;doa<N_DOAS;doa++)   //Over present estimated list
	{
	if(N_SOURCE<8)
	  {
          SOURCE[N_SOURCE].DOA=DOA_list[doa];   //Start a new data structure for that source.
	  SOURCE[N_SOURCE].DOA_SET=1;
          SOURCE[N_SOURCE].sig=(float *)calloc(LENGTH,sizeof(float));                //Initialize  the signal part as empty.
	  SOURCE[N_SOURCE].L_src=0;             //Length of source signal after initialising.
	  SOURCE[N_SOURCE].L_init=(l+1)-L;    //Length of source signal after initialising.
          SOURCE[N_SOURCE].WT=0;                //Initialise the weighting/flag for inclusion in this cycle. Mainly to determine if active block.
          SOURCE[N_SOURCE].WTDOA=1;            // Initialise the weighting/flag for inclusion in this cycle. Mainly to determine if DOA is set.
	  N_SOURCE++;				//Incrementing source count
	  }
	}

      //Removing reduntant sources.
      N_SOURCE_tmp=N_SOURCE;
      N_SOURCE=0;
      for(src=0;src<N_SOURCE_tmp;src++)
	{
        if (((SOURCE[src].WT<6)&(SOURCE[src].WTDOA<5)) || (SOURCE[src].L_src > (L*6)))  //If recently active or relatively old, keep the source (VARIABLE)
	  {
          SOURCE_tmp[N_SOURCE]=SOURCE[src];
          N_SOURCE++;

       	  SOURCE[src].DOA_SET=0;
	  SOURCE[src].L_src=0;             //Length of source signal after initialising.
	  SOURCE[src].L_init=0;    //Length of source signal after initialising.
          SOURCE[src].WT=0;                //Reseting to 0.
          SOURCE[src].WTDOA=0;            // Reseting to 0.
	  }
	else          //Else free the source
	  {
	  free(SOURCE[src].sig);
	  SOURCE[src].sig=NULL;
	  SOURCE[src].DOA_SET=0;
	  SOURCE[src].L_src=0;             //Length of source signal after initialising.
	  SOURCE[src].L_init=0;    //Length of source signal after initialising.
          SOURCE[src].WT=0;                //Reseting to 0.
          SOURCE[src].WTDOA=0;            // Reseting to 0.
	  }
	}
      for(src=0;src<N_SOURCE;src++)       //Implementing SOURCE=SOURCE_tmp;
	{
        SOURCE[src]=SOURCE_tmp[src];
        }
      
        //RECOVERY PART
      for(q=0;q<N_SOURCE;q++)   //Over sources
	{
	  //Implementing  WEIGHT_SP(q,:)=(abs(DOA_full-SOURCE{q}.DOA)<10); 
	WEIGHT_SP_SUM[q]=0;    //For summing weights
	for(m=0;m<(L+L_pad);m++)
	  WEIGHT_SP[q][m]=0;
	for (m=1;m<(L_DOA_end);m++)
	  {
          WEIGHT_SP[q][m]=(fabsf(DOA_full[m]-SOURCE[q].DOA)<20) ? 1:0;        //Weighting for sources spectrum using DOA information. (VARIABLE)   was 10 on 11-11-2003
	  WEIGHT_SP[q][m-1]=(WEIGHT_SP[q][m]==1) ? 1:0;         //For WEIGHT_SP(q,:)= (WEIGHT_SP(q,:)+[WEIGHT_SP(q,2:end) 0]+[0 WEIGHT_SP(q,1:end-1)])>0;   To create a window around the set frequency point
	  WEIGHT_SP[q][m+1]=((WEIGHT_SP[q][m]==1) && (WEIGHT_SP[q][m+1]==0)) ? 2:WEIGHT_SP[q][m+1]; // ''
	  }       //To create a window around the set frequency point
	for (m=1;m<(L_DOA_end)-1;m++) //For WEIGHT_SP(q,:)= (WEIGHT_SP(q,:)+[WEIGHT_SP(q,2:end) 0]+[0 WEIGHT_SP(q,1:end-1)])>0;
	  {
	  WEIGHT_SP[q][m]=(WEIGHT_SP[q][m]==2) ? 1:WEIGHT_SP[q][m];
	  WEIGHT_SP[q][(L+L_pad-1)-m]=WEIGHT_SP[q][m];
	  if(WEIGHT_SP[q][m]>0)
	    {
            WEIGHT_SP_SUM[q]+=2;
	    }
	  }
        } 

      for(q=0;q<N_SOURCE;q++)   //Over sources
	{
        for(m=0;m< (L+L_pad);m++)   //Initiallising weights for using pitch info. Implementing WGHT_TMP=zeros(size(DOA_full));
          {
          WGHT_TMP[m]=0;
          }
        for(m=0;m<=(L+L_pad)/2;m++)           // Corrected
	  {
	  WGHT_TMP[(L+L_pad-1)-m]=WGHT_TMP[m];  //Symmetrising;   WGHT_TMP=WGHT_TMP+[0 fliplr(WGHT_TMP(2:end))]; 
	  WEIGHT_SP[q][m]= ((WEIGHT_SP[q][m]==1) || (WGHT_TMP[m]==1)) ? 1:0;        //Updating the weighting for qth source using pitch info
          WEIGHT_SP[q][(L+L_pad-1)-m]=WEIGHT_SP[q][m];          //Updating the weighting for qth source using pitch info
	  }
        }


       //Recovering from the output of the filterbank
      for(q=0;q<N_SOURCE;q++)
	{
	SOURCE[q].L_src+=L;
        for(i=1;i<=((L+L_pad)/2);i++)
	  {
	  if(WEIGHT_SP[q][i]>0)
	    {
	    for(m=0;m<L;m++)
	      {
              if(SOURCE[q].DOA<75)       // Added on 6-10
                SOURCE[q].sig[(l+1-L)+m]+=2*FB_ORIG[m][1][i].re/(L+L_pad);
              else                       // Added on 6-10
              if(SOURCE[q].DOA>105)      // Added on 6-10
                SOURCE[q].sig[(l+1-L)+m]+=2*FB_ORIG[m][0][i].re/(L+L_pad); // Added on 6-10
              else                                                         // Added on 6-10
                SOURCE[q].sig[(l+1-L)+m]+=(FB_ORIG[m][1][i].re/(L+L_pad)+FB_ORIG[m][0][i].re/(L+L_pad)); // Added on 6-10
	      }
	    }
	  }
	}
      printf("l=%d\n",l);
      free(DOA_list);
      DOA_list=NULL;
      free(ID_PK);
      ID_PK=NULL;
      }
    }
  
   //Freeing FB_ORIG 
  for(ccnt =0;ccnt<L;ccnt++)
    {
    for(rcnt = 0;rcnt<M; rcnt++)
      free(FB_ORIG[ccnt][rcnt]);       
    free(FB_ORIG[ccnt]);
    }
  for(m=0;m<SOURCE_NUM;m++)
    {
    free(WEIGHT_SP[m]);
    }
  free(WEIGHT_SP);
  free(DOA_AT_MAP);
  DOA_AT_MAP=NULL;
  free(SPEC1);
  SPEC1=NULL;
  free(SPEC2);
  SPEC2=NULL;
  free(SPEC);
  SPEC=NULL;
  free(WGHT_TMP);
  WGHT_TMP=NULL;
  free(DOA_full);
  DOA_full=NULL;
  free(FB_ORIG); 
  FB_ORIG=NULL; 
  free(DOAs); 
  DOAs=NULL;
  free(DOA_wt_hist);
  DOA_wt_hist=NULL;
  free(DOA_hist_idx);
  DOA_hist_idx=NULL;
  fftw_destroy_plan(FPLN); 
  fftw_free(TMP1); fftw_free(TMP2);        //Clearing temporary arrays used by fftw
  source_display_window_label=(char *)calloc(300,sizeof(char));
  sprintf(source_display_window_label,"source diaply (sigsep_final) on ");
  source_display_window_label=strcat(source_display_window_label,Data.file_name); //Creating new window label
  m=source_plot_init(source_display_window_label);                    //Calling function to plot the separated sources.
  /*source_plot();*/
  }

